﻿using System;
using System.Collections.Concurrent;
using System.Diagnostics;
using System.IO.MemoryMappedFiles;
using System.Threading;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Diagnostics.Runtime;


namespace IpcMMF
{
    public class IpcMmfTransition : IDisposable
    {
        const int _initializePreamble = 0xAABBCC;

        const int _SKEPTICAL_LAZY_COUNT = 500;
        const int _MAX_PARTICIPANTS = 10;
        const int _THRED_ID_BYTES = _MAX_PARTICIPANTS * 4; // 10*4bytes
        const int _PROCESS_ID_BYTES = _MAX_PARTICIPANTS * 4; // 10*4bytes

        const long _POS_PEAMBLE = 0;
        const long _POS_PARTICIPANTS = 4;
        const long _POS_PARTICIPANTS_THREADIDS = 5; // 10* 4byte
        const long _POS_PARTICIPANTS_PROCESSIDS = 5 + _THRED_ID_BYTES; // 10* 4byte
        const long _POS_READBYPARTICIPANTS = 5 + _THRED_ID_BYTES + _PROCESS_ID_BYTES;
        const long _POS_MESSAGENUMBER = 6 + _THRED_ID_BYTES + _PROCESS_ID_BYTES;
        const long _POS_DATACOUNT = 10 + _THRED_ID_BYTES + _PROCESS_ID_BYTES;
        const long _POS_MSGPARTFROM = 14 + _THRED_ID_BYTES + _PROCESS_ID_BYTES;
        const long _POS_MSGPARTTO = 18 + _THRED_ID_BYTES + _PROCESS_ID_BYTES;
        const long _POS_MSGTOTAL = 22 + _THRED_ID_BYTES + _PROCESS_ID_BYTES;
        const long _POS_DATASTART = 26 + _THRED_ID_BYTES + _PROCESS_ID_BYTES;

        MemoryMappedViewAccessor _acessor;
        ConcurrentQueue<byte[]> _messagesToSend;
        Mutex _mmfMutex;
        MemoryMappedFile _sharedMemory;

        int _didNothingDuringATransmissionCount = 0;
        int _myParticipantIndex;
        int _allreadyTransmitted;
        bool _communicationRequiered;
        Thread _communicationThread;
        byte[] _currentMessage;
        byte[] _incommingData;
        int _lastReadOrWriteMessage;

        long _mmfSize;
        int _restTime = 1;
        bool _useOwnThread;

        public IpcMmfTransition(string mapName, bool useOwnThread = true, long fileSize = 10 * 1024 * 1024)  // 10 Mb cash default!
        {
            _mmfSize = Math.Max(fileSize, 100);
            _useOwnThread = useOwnThread;
            _myParticipantIndex = -1;
            _messagesToSend = new ConcurrentQueue<byte[]>();
            _currentMessage = null;
            _allreadyTransmitted = 0;
            _incommingData = null;
            _lastReadOrWriteMessage = 0;
            _sharedMemory = MemoryMappedFile.CreateOrOpen(mapName, _mmfSize);
            _acessor = _sharedMemory.CreateViewAccessor();

            _mmfMutex = new Mutex(false, mapName + "Mutex");

            InitializeMemoryMapedFile();

            // Start Communication Thread
            _communicationRequiered = true;
            if (_useOwnThread)
                _communicationThread = new Thread(CommunicationThread);
        }
        public bool FileWasAllreadyOpend { get; private set; }
        public void Dispose()
        {
            _communicationRequiered = false;
            JoinOrLeft(true);

            // Wait for the thread to close
            if (_communicationThread != null)
                while (_communicationThread.IsAlive)
                    Thread.Sleep(1);

            _acessor.Dispose();
            _sharedMemory.Dispose();

            _acessor = null;
            _sharedMemory = null;
            _messagesToSend = null;
        }

        public void Run()
        {
            if (_useOwnThread)
                _communicationThread.Start();

            else
                CommunicationThread();
        }

        public event Action<IpcMmfTransition, byte[]> DataAvailable;
        public event Action<IpcMmfTransition> ParticipantLeft;

        void InitializeMemoryMapedFile()
        {
            _mmfMutex.WaitOne();

            int p = Preamble;

            // Is Initialized?
            if (p != _initializePreamble)
            {
                // No.. so do init...
                Preamble = _initializePreamble;
                Participants = 0;
                for (int i = 0; i < _MAX_PARTICIPANTS; i++)
                {
                    WriteThreadId(i, 0);
                    WriteProcessId(i, 0);
                }
                ReadByParticipants = 0;
                DataCount = 0;
                MsgPartFrom = 0;
                MsgPartTo = 0;
                MsgTotal = 0;
                FileWasAllreadyOpend = false;
            }
            else
                FileWasAllreadyOpend = true;

            _mmfMutex.ReleaseMutex();
        }

        /// <summary>
        /// Joins or left the memory mapped file.
        /// To correct join a communication, this function should be called in the communication thread in order to
        /// register the correct managed thread id.
        /// </summary>
        /// <param name="orLeft"></param>
        void JoinOrLeft(bool orLeft = false)
        {
            bool joined = false;

            while (!joined)
            {
                // Wait until there is not transmissiong on the stream
                _mmfMutex.WaitOne();

                if (!TransmisstionRunning)
                {
                    // Increment Participants and ReadByParticipants
                    // in order to mark that there is not transmittion.
                    if (orLeft)
                    {
                        DecrementParticipants();

                        // Also decrement ReadByParticipants
                        ReadByParticipants = (byte)(ReadByParticipants - 1);

                        // Reset thread and process id
                        if (_myParticipantIndex >= 0)
                        {
                            WriteThreadId(_myParticipantIndex, 0);
                            WriteProcessId(_myParticipantIndex, 0);
                        }
                    }
                    else
                    {
                        if (Participants >= (_MAX_PARTICIPANTS - 1))
                            throw new Exception("Number of participants exceeded");

                        // Increment Participants and ReadByParticipants
                        Participants = (byte)(Participants + 1);
                        ReadByParticipants = (byte)(ReadByParticipants + 1);

                        // Find free thread and process slot
                        for (int i = 0; i < _MAX_PARTICIPANTS; i++)
                            if (ReadThreadId(i) == 0)
                            {
                                _myParticipantIndex = i;
                                break;
                            }

                        // Register Thread and Process ID
                        WriteThreadId(_myParticipantIndex, Thread.CurrentThread.ManagedThreadId);
                        WriteProcessId(_myParticipantIndex, Process.GetCurrentProcess().Id);
                    }
                    joined = true;
                }

                _mmfMutex.ReleaseMutex();
            }
        }

        void CommunicationThread()
        {
            // Already jointed? if not, do so!
            if (_myParticipantIndex < 0)
                JoinOrLeft();

            while (_communicationRequiered)
            {
                bool didSomething = false;

                try
                {
                    _mmfMutex.WaitOne();

                    // Is there no transmission running?
                    if (!TransmisstionRunning)
                    {
                        // _incommingData will be null if no more data is expected.
                        // If _incommingData is not null, do nothing and wait for the responsible thread
                        // to finish the transmission. (Or this is my transmission)
                        if (_incommingData == null || _currentMessage != null)
                            // Is there an unfinished transmision?
                            if (_currentMessage != null)
                            {
                                ProcessSend();
                                didSomething = true;
                            }
                            // Do i have something new to send?
                            else if (_messagesToSend.Count > 0)
                            {
                                byte[] message;
                                if (_messagesToSend.TryDequeue(out message))
                                {
                                    _currentMessage = message;
                                    _allreadyTransmitted = 0;
                                    ProcessSend();
                                    didSomething = true;
                                }
                            }
                    }
                    else
                    {
                        // Incomming Data
                        if (_currentMessage == null) // We dont have to read our own message
                            didSomething = Read();

                        if (!didSomething)
                            _didNothingDuringATransmissionCount++;
                    }

                    if (!didSomething && _didNothingDuringATransmissionCount > _SKEPTICAL_LAZY_COUNT)
                        DidNothingForTooLongDuringATransmission();
                }
                catch (AbandonedMutexException ex)
                {
                    // Decrement 
                    DecrementParticipants();
                }
                finally
                {
                    _mmfMutex.ReleaseMutex();
                }

                if (!didSomething)
                {
                    // I am done for now... lets get some rest
                    Thread.Sleep(_restTime);
                }
                else
                    _didNothingDuringATransmissionCount = 0;
            }
        }

        void DecrementParticipants()
        {
            Participants = (byte)(Participants - 1);
            if (ParticipantLeft != null)
                ParticipantLeft(this);
        }

        /// <summary>
        /// Time out error handler for a transmission.

        /// </summary>
        void DidNothingForTooLongDuringATransmission()
        {
            // Get All process available id´s
            List<int> processIds = new List<int>();
            for (int i = 0; i < _MAX_PARTICIPANTS; i++)
            {
                int pid = ReadProcessId(i);
                if (pid != 0)
                    processIds.Add(pid);
            }

            // Check if the communication is inside one process 
            int firstProcessId = processIds.FirstOrDefault();
            bool allSameProcess = processIds.All(t => t == firstProcessId);

            if (allSameProcess)
            {
                // The communication is between one process, so check if all registered threads are running.
                List<int> threadIds = new List<int>();
                for (int i = 0; i < _MAX_PARTICIPANTS; i++)
                {
                    int tid = ReadThreadId(i);
                    if (tid != 0)
                        threadIds.Add(tid);
                }

                // Get all running threads by creating a debug dump.
                try
                {
                    using (DataTarget target = DataTarget.AttachToProcess(Process.GetCurrentProcess().Id, 1000, AttachFlag.Passive))
                    {
                        ClrRuntime runtime = target.ClrVersions.First().CreateRuntime();
                        foreach (ClrThread thread in runtime.Threads)
                        {
                            if (threadIds.Contains(thread.ManagedThreadId))
                                if (thread.IsAlive)
                                    threadIds.Remove(thread.ManagedThreadId);
                        }
                    }
                }
                catch (Exception)
                {
                    // Try again later
                    return;
                }

                // All remaining ids relate to dead threats
                if (threadIds.Count > 0)
                {
                    ErrorHandlingDecrementParticipants(threadIds.Count);

                    for (int i = 0; i < _MAX_PARTICIPANTS; i++)
                    {
                        int tid = ReadThreadId(i);
                        if (threadIds.Contains(tid))
                        {
                            WriteThreadId(i, 0);
                            WriteProcessId(i, 0);
                        }
                    }
                }
            }
            else
            {
                // Check if all processes are running
                foreach (var p in Process.GetProcesses())
                    while (processIds.Contains(p.Id))
                        processIds.Remove(p.Id);

                if (processIds.Count > 0)
                {
                    ErrorHandlingDecrementParticipants(processIds.Count);

                    for (int i = 0; i < _MAX_PARTICIPANTS; i++)
                    {
                        int pid = ReadProcessId(i);
                        if (processIds.Contains(pid))
                        {
                            WriteThreadId(i, 0);
                            WriteProcessId(i, 0);
                        }
                    }
                }
            }

            _didNothingDuringATransmissionCount = 0;
        }

        void ErrorHandlingDecrementParticipants(int by)
        {
            for (int i = 0; i < by; i++)
                DecrementParticipants();
        }

        void ProcessSend()
        {
            _allreadyTransmitted += PostData(_currentMessage, _allreadyTransmitted);
            if (_allreadyTransmitted >= _currentMessage.Length)
            {
                // We are done...
                _currentMessage = null;
                _allreadyTransmitted = 0;
            }
        }
        int PostData(byte[] message, int allreadyTransmitted)
        {
            int bytesToSend = Math.Min(message.Length - allreadyTransmitted, (int)(_mmfSize - _POS_DATASTART));

            ReadByParticipants = 0;
            MessageNumber = MessageNumber + 1;
            _lastReadOrWriteMessage = MessageNumber;
            DataCount = bytesToSend;
            MsgPartFrom = Math.Max(0, allreadyTransmitted);
            MsgPartTo = MsgPartFrom + bytesToSend;
            MsgTotal = message.Length;

            _acessor.WriteArray(_POS_DATASTART, message, allreadyTransmitted, bytesToSend);

            return bytesToSend;
        }

        /// <summary>
        /// Reads the current message from the mmf.
        /// </summary>
        /// <return>True if the message was read, false if the message already have been read.</returns>
        bool Read()
        {
            // Check if i read that message?
            if (_lastReadOrWriteMessage != MessageNumber)
            {
                _lastReadOrWriteMessage = MessageNumber;
                if (_incommingData == null)
                    _incommingData = new byte[MsgTotal];
                _acessor.ReadArray(_POS_DATASTART, _incommingData, MsgPartFrom, DataCount);

                // Increment read counter
                ReadByParticipants = (byte)(ReadByParticipants + 1);

                // End of Message?
                if (MsgPartTo >= MsgTotal)
                {
                    if (DataAvailable != null)
                        DataAvailable(this, _incommingData);
                    _incommingData = null;
                }
                return true;
            }
            return false;
        }

        /// <summary>
        /// Send data to all Participants
        /// </summary>
        /// <param name="data"></param>
        public void Transmit(byte[] data)
        {
            _messagesToSend.Enqueue(data);
        }

        public int GetNumberOfParticipants()
        {
            int p;
            _mmfMutex.WaitOne();
            p = Participants;
            _mmfMutex.ReleaseMutex();
            return p;
        }


        #region MMF Access Functions

        bool TransmisstionRunning
        {
            get { return ReadByParticipants < Participants - 1; }
        }

        int Preamble
        {
            get { return _acessor.ReadInt32(_POS_PEAMBLE); }
            set { _acessor.Write(_POS_PEAMBLE, value); }
        }
        byte Participants
        {
            get { return _acessor.ReadByte(_POS_PARTICIPANTS); }
            set { _acessor.Write(_POS_PARTICIPANTS, value); }
        }

        int ReadThreadId(int participantIndex)
        {
            return _acessor.ReadInt32(_POS_PARTICIPANTS_THREADIDS + 4 * participantIndex);
        }
        void WriteThreadId(int participantIndex, int id)
        {
            _acessor.Write(_POS_PARTICIPANTS_THREADIDS + participantIndex * 4, id);
        }

        int ReadProcessId(int participantIndex)
        {
            return _acessor.ReadInt32(_POS_PARTICIPANTS_PROCESSIDS + 4 * participantIndex);
        }
        void WriteProcessId(int participantIndex, int id)
        {
            _acessor.Write(_POS_PARTICIPANTS_PROCESSIDS + participantIndex * 4, id);
        }

        byte ReadByParticipants
        {
            get { return _acessor.ReadByte(_POS_READBYPARTICIPANTS); }
            set { _acessor.Write(_POS_READBYPARTICIPANTS, value); }
        }
        int MessageNumber // To check if a perticipant allready read this message
        {
            get { return _acessor.ReadByte(_POS_MESSAGENUMBER); }
            set { _acessor.Write(_POS_MESSAGENUMBER, value); }
        }
        int DataCount
        {
            get { return _acessor.ReadInt32(_POS_DATACOUNT); }
            set { _acessor.Write(_POS_DATACOUNT, value); }
        }
        int MsgPartFrom
        {
            get { return _acessor.ReadInt32(_POS_MSGPARTFROM); }
            set { _acessor.Write(_POS_MSGPARTFROM, value); }
        }
        int MsgPartTo
        {
            get { return _acessor.ReadInt32(_POS_MSGPARTTO); }
            set { _acessor.Write(_POS_MSGPARTTO, value); }
        }
        int MsgTotal
        {
            get { return _acessor.ReadInt32(_POS_MSGTOTAL); }
            set { _acessor.Write(_POS_MSGTOTAL, value); }
        }

        # endregion
    }

    class Program
    {
        static void Main(string[] args)
        {
            IpcMmfTransition participantA = new IpcMmfTransition("TestChannel");
            IpcMmfTransition participantB = new IpcMmfTransition("TestChannel");
            IpcMmfTransition participantC = new IpcMmfTransition("TestChannel");

            participantA.DataAvailable += (transition, bytes) =>
            {
                Console.WriteLine("A received: " + Encoding.ASCII.GetString(bytes));
            };
            participantB.DataAvailable += (transition, bytes) =>
            {
                Console.WriteLine("B received: " + Encoding.ASCII.GetString(bytes));
            };
            participantC.DataAvailable += (transition, bytes) =>
            {
                Console.WriteLine("C received: " + Encoding.ASCII.GetString(bytes));
            };

            participantA.Run();
            participantB.Run();
            participantC.Run();

            while(participantC.GetNumberOfParticipants() < 3)
                Thread.Sleep(1);

            participantA.Transmit(Encoding.ASCII.GetBytes("A says Hello"));
            participantB.Transmit(Encoding.ASCII.GetBytes("B says Hello"));
            participantC.Transmit(Encoding.ASCII.GetBytes("C says Hello"));

            Thread.Sleep(500);

            participantA.Dispose();
            participantB.Dispose();
            participantC.Dispose();

            Console.ReadKey();
        }
    }
}
